@@ -2,7 +2,7 @@
  *
  *  Wireless daemon for Linux
  *
- *  Copyright (C) 2013-2014  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2013-2019  Intel Corporation. All rights reserved.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -38,31 +38,25 @@
 #include "src/dbus.h"
 #include "src/eap.h"
 #include "src/eapol.h"
-#include "src/scan.h"
 #include "src/rfkill.h"
-#include "src/ap.h"
 #include "src/plugin.h"
-#include "src/simauth.h"
-#include "src/adhoc.h"
-#include "src/blacklist.h"
 #include "src/storage.h"
-#include "src/erp.h"
+#include "src/anqp.h"
 
 #include "src/backtrace.h"
 
 static struct l_genl *genl;
-static struct l_genl_family *nl80211;
 static struct l_settings *iwd_config;
 static struct l_timeout *timeout;
 static const char *interfaces;
 static const char *nointerfaces;
 static const char *phys;
 static const char *nophys;
-static const char *config_dir;
 static const char *plugins;
 static const char *noplugins;
 static const char *debugopt;
 static bool terminating;
+static bool nl80211_complete;
 
 static void main_loop_quit(struct l_timeout *timeout, void *user_data)
 {
@@ -76,7 +70,7 @@ static void iwd_shutdown(void)
 
 	terminating = true;
 
-	if (!nl80211) {
+	if (!nl80211_complete) {
 		l_main_quit();
 		return;
 	}
@@ -103,6 +97,31 @@ const struct l_settings *iwd_get_config(void)
 	return iwd_config;
 }
 
+struct l_genl *iwd_get_genl(void)
+{
+	return genl;
+}
+
+const char *iwd_get_iface_whitelist(void)
+{
+	return interfaces;
+}
+
+const char *iwd_get_iface_blacklist(void)
+{
+	return nointerfaces;
+}
+
+const char *iwd_get_phy_whitelist(void)
+{
+	return phys;
+}
+
+const char *iwd_get_phy_blacklist(void)
+{
+	return nophys;
+}
+
 static void usage(void)
 {
 	printf("iwd - Wireless daemon\n"
@@ -114,10 +133,10 @@ static void usage(void)
 		"\t-I, --nointerfaces     Interfaces to ignore\n"
 		"\t-p, --phys             Phys to manage\n"
 		"\t-P, --nophys           Phys to ignore\n"
-		"\t-c, --config           Configuration directory to use\n"
 		"\t-l, --plugin           Plugins to include\n"
 		"\t-L, --noplugin         Plugins to exclude\n"
 		"\t-d, --debug            Enable debug output\n"
+		"\t-v, --version          Show version\n"
 		"\t-h, --help             Show help options\n");
 }
 
@@ -128,7 +147,6 @@ static const struct option main_options[] = {
 	{ "nointerfaces", required_argument, NULL, 'I' },
 	{ "phys",         required_argument, NULL, 'p' },
 	{ "nophys",       required_argument, NULL, 'P' },
-	{ "config",       required_argument, NULL, 'c' },
 	{ "plugin",       required_argument, NULL, 'l' },
 	{ "noplugin",     required_argument, NULL, 'L' },
 	{ "debug",        optional_argument, NULL, 'd' },
@@ -143,38 +161,19 @@ static void do_debug(const char *str, void *user_data)
 	l_info("%s%s", prefix, str);
 }
 
-static void nl80211_appeared(void *user_data)
+static void nl80211_appeared(const struct l_genl_family_info *info,
+							void *user_data)
 {
-	struct l_genl_family *nl80211 = user_data;
-
-	if (terminating)
-		return;
-
 	l_debug("Found nl80211 interface");
 
-	manager_init(nl80211, interfaces, nointerfaces);
+	nl80211_complete = true;
 
-	if (!wiphy_init(nl80211, phys, nophys))
-		l_error("Unable to init wiphy functionality");
+	if (iwd_modules_init() < 0) {
+		l_main_quit();
+		return;
+	}
 
-	netdev_set_nl80211(nl80211);
-
-	if (!scan_init(nl80211))
-		l_error("Unable to init scan functionality");
-
-	ap_init(nl80211);
-	adhoc_init(nl80211);
-}
-
-static void nl80211_vanished(void *user_data)
-{
-	l_debug("Lost nl80211 interface");
-
-	manager_exit();
-	ap_exit();
-	adhoc_exit();
-	scan_exit();
-	wiphy_exit();
+	plugin_init(plugins, noplugins);
 }
 
 static void request_name_callback(struct l_dbus *dbus, bool success,
@@ -185,26 +184,12 @@ static void request_name_callback(struct l_dbus *dbus, bool success,
 		goto fail_exit;
 	}
 
-	if (!l_dbus_object_manager_enable(dbus))
+	if (!l_dbus_object_manager_enable(dbus, "/"))
 		l_warn("Unable to register the ObjectManager");
 
-	genl = l_genl_new_default();
-	if (!genl) {
-		l_error("Failed to open generic netlink socket");
-		goto fail_exit;
-	}
-
-	if (getenv("IWD_GENL_DEBUG"))
-		l_genl_set_debug(genl, do_debug, "[GENL] ", NULL);
-
-	nl80211 = l_genl_family_new(genl, NL80211_GENL_NAME);
-	if (!nl80211) {
-		l_error("Failed to open nl80211 interface");
-		goto fail_exit;
-	}
-
-	l_genl_family_set_watches(nl80211, nl80211_appeared, nl80211_vanished,
-								nl80211, NULL);
+	/* TODO: Always request nl80211 for now, ignoring auto-loading */
+	l_genl_request_family(genl, NL80211_GENL_NAME, nl80211_appeared,
+				NULL, NULL);
 	return;
 
 fail_exit:
@@ -215,7 +200,7 @@ static void dbus_ready(void *user_data)
 {
 	struct l_dbus *dbus = user_data;
 
-	l_dbus_name_acquire(dbus, "net.connman.iwd", false, false, true,
+	l_dbus_name_acquire(dbus, "net.connman.iwd", false, false, false,
 				request_name_callback, NULL);
 }
 
@@ -380,13 +365,14 @@ int main(int argc, char *argv[])
 	bool enable_dbus_debug = false;
 	int exit_status;
 	struct l_dbus *dbus;
-	char *config_path;
-	uint32_t eap_mtu;
+	const char *config_dir;
+	char **config_dirs;
+	int i;
 
 	for (;;) {
 		int opt;
 
-		opt = getopt_long(argc, argv, "Bi:I:p:P:c:d::vh",
+		opt = getopt_long(argc, argv, "Bi:I:p:P:d::vh",
 							main_options, NULL);
 		if (opt < 0)
 			break;
@@ -407,9 +393,6 @@ int main(int argc, char *argv[])
 		case 'P':
 			nophys = optarg;
 			break;
-		case 'c':
-			config_dir = optarg;
-			break;
 		case 'l':
 			plugins = optarg;
 			break;
@@ -457,29 +440,46 @@ int main(int argc, char *argv[])
 
 	l_info("Wireless daemon version %s", VERSION);
 
+	config_dir = getenv("CONFIGURATION_DIRECTORY");
 	if (!config_dir)
 		config_dir = DAEMON_CONFIGDIR;
 
-	config_path = l_strdup_printf("%s/%s", config_dir, "main.conf");
+	l_debug("Using configuration directory %s", config_dir);
+
 	iwd_config = l_settings_new();
 
-	if (!l_settings_load_from_file(iwd_config, config_path))
-		l_warn("Skipping optional configuration file %s", config_path);
+	config_dirs = l_strsplit(config_dir, ':');
+	for (i = 0; config_dirs[i]; i++) {
+		char *path = l_strdup_printf("%s/%s", config_dirs[i],
+								"main.conf");
+		bool result = l_settings_load_from_file(iwd_config, path);
+		l_free(path);
 
-	l_free(config_path);
+		if (result) {
+			l_info("Loaded configuration from %s/main.conf",
+							config_dirs[i]);
+			break;
+		}
+	}
+	l_strv_free(config_dirs);
 
 	__eapol_set_config(iwd_config);
-
-	if (!l_settings_get_uint(iwd_config, "EAP", "mtu", &eap_mtu))
-		eap_mtu = 1400; /* on WiFi the real MTU is around 2304 */
+	__eap_set_config(iwd_config);
 
 	exit_status = EXIT_FAILURE;
 
-	if (create_dirs(DAEMON_STORAGEDIR "/")) {
-		l_error("Failed to create " DAEMON_STORAGEDIR "/");
+	if (!storage_create_dirs())
 		goto fail_dbus;
+
+	genl = l_genl_new();
+	if (!genl) {
+		l_error("Failed to open generic netlink socket");
+		goto fail_genl;
 	}
 
+	if (getenv("IWD_GENL_DEBUG"))
+		l_genl_set_debug(genl, do_debug, "[GENL] ", NULL);
+
 	dbus = l_dbus_new_default(L_DBUS_SYSTEM_BUS);
 	if (!dbus) {
 		l_error("Failed to initialize D-Bus");
@@ -493,48 +493,17 @@ int main(int argc, char *argv[])
 	l_dbus_set_disconnect_handler(dbus, dbus_disconnected, NULL, NULL);
 	dbus_init(dbus);
 
-	eap_init(eap_mtu);
-	eapol_init();
-	rfkill_init();
-
-	if (!netdev_init(interfaces, nointerfaces))
-		goto fail_netdev;
-
-	if (!device_init())
-		goto fail_device;
-
-	station_init();
-	wsc_init();
-	network_init();
-	known_networks_init();
-	sim_auth_init();
-	plugin_init(plugins, noplugins);
-	blacklist_init();
-	erp_init();
-
 	exit_status = l_main_run_with_signal(signal_handler, NULL);
-
-	erp_exit();
-	blacklist_exit();
 	plugin_exit();
-	sim_auth_exit();
-	known_networks_exit();
-	network_exit();
-	wsc_exit();
-	station_exit();
-	device_exit();
-fail_device:
-	netdev_exit();
-fail_netdev:
-	rfkill_exit();
-	eapol_exit();
-	eap_exit();
 
-	l_genl_family_unref(nl80211);
-	l_genl_unref(genl);
+	iwd_modules_exit();
+
 	dbus_exit();
 	l_dbus_destroy(dbus);
+	storage_cleanup_dirs();
 fail_dbus:
+	l_genl_unref(genl);
+fail_genl:
 	l_settings_free(iwd_config);
 
 	l_timeout_remove(timeout);
