@@ -2,7 +2,7 @@
  *
  *  Wireless daemon for Linux
  *
- *  Copyright (C) 2013-2015  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2013-2019  Intel Corporation. All rights reserved.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -48,8 +48,13 @@
 #include "src/knownnetworks.h"
 #include "src/network.h"
 #include "src/blacklist.h"
+#include "src/util.h"
+
+static uint32_t known_networks_watch;
 
 struct network {
+	char ssid[33];
+	enum security security;
 	char *object_path;
 	struct station *station;
 	struct network_info *info;
@@ -60,20 +65,22 @@ struct network {
 	struct l_settings *settings;
 	struct l_queue *secrets;
 	struct l_queue *blacklist; /* temporary blacklist for BSS's */
+	uint8_t hessid[6];
+	char **nai_realms;
+	uint8_t *rc_ie;
 	bool update_psk:1;  /* Whether PSK should be written to storage */
 	bool ask_passphrase:1; /* Whether we should force-ask agent */
+	bool is_hs20:1;
 	int rank;
 };
 
-static struct l_queue *networks;
-
 static bool network_settings_load(struct network *network)
 {
 	if (network->settings)
 		return true;
 
-	network->settings = storage_network_open(network_get_security(network),
-							network->info->ssid);
+	if (network->info)
+		network->settings = network_info_open_settings(network->info);
 
 	return network->settings != NULL;
 }
@@ -109,20 +116,6 @@ static void network_settings_close(struct network *network)
 	network->settings = NULL;
 }
 
-bool network_info_match(const void *a, const void *b)
-{
-	const struct network_info *ni_a = a;
-	const struct network_info *ni_b = b;
-
-	if (ni_a->type != ni_b->type)
-		return false;
-
-	if (strcmp(ni_a->ssid, ni_b->ssid))
-		return false;
-
-	return true;
-}
-
 static bool network_secret_check_cacheable(void *data, void *user_data)
 {
 	struct eap_secret_info *secret = data;
@@ -137,18 +130,11 @@ static bool network_secret_check_cacheable(void *data, void *user_data)
 
 void network_connected(struct network *network)
 {
+	enum security security = network_get_security(network);
+	const char *ssid = network_get_ssid(network);
 	int err;
 
-	/*
-	 * This triggers an update to network->info->connected_time and
-	 * other possible actions in knownnetworks.c.
-	 */
-	err = storage_network_touch(network_get_security(network),
-					network->info->ssid);
-	switch (err) {
-	case 0:
-		break;
-	case -ENOENT:
+	if (!network->info) {
 		/*
 		 * This is an open network seen for the first time:
 		 *
@@ -159,13 +145,14 @@ void network_connected(struct network *network)
 		if (!network->settings)
 			network->settings = l_settings_new();
 
-		storage_network_sync(network_get_security(network),
-					network->info->ssid,
-					network->settings);
-		break;
-	default:
-		l_error("Error %i touching network config", err);
-		break;
+		storage_network_sync(security, ssid, network->settings);
+	} else {
+		err = network_info_touch(network->info);
+		if (err < 0)
+			l_error("Error %i touching network config", err);
+
+		/* Syncs frequencies of already known network*/
+		known_network_frequency_sync(network->info);
 	}
 
 	l_queue_foreach_remove(network->secrets,
@@ -179,35 +166,6 @@ void network_disconnected(struct network *network)
 	network_settings_close(network);
 }
 
-struct network_find_rank_data {
-	const struct network_info *info;
-	int n;
-};
-
-static bool network_find_rank_update(const struct network_info *network,
-					void *user_data)
-{
-	struct network_find_rank_data *data = user_data;
-
-	if (network == data->info)
-		return false;
-
-	if (network->seen_count)
-		data->n++;
-
-	return true;
-}
-
-static int network_find_rank_index(const struct network_info *info)
-{
-	struct network_find_rank_data data = { info, 0 };
-
-	if (!known_networks_foreach(network_find_rank_update, &data))
-		return data.n;
-
-	return -1;
-}
-
 /* First 64 entries calculated by 1 / pow(n, 0.3) for n >= 1 */
 static const double rankmod_table[] = {
 	1.0000000000, 0.8122523964, 0.7192230933, 0.6597539554,
@@ -238,11 +196,11 @@ bool network_rankmod(const struct network *network, double *rankmod)
 	 * to at least once are autoconnectable.  Known Networks that
 	 * we have never connected to are not.
 	 */
-	if (!network->info->connected_time.tv_sec)
+	if (!network->info || !network->info->connected_time)
 		return false;
 
-	n = network_find_rank_index(network->info);
-	if (n == -1)
+	n = known_network_offset(network->info);
+	if (n < 0)
 		return false;
 
 	nmax = L_ARRAY_SIZE(rankmod_table);
@@ -255,54 +213,6 @@ bool network_rankmod(const struct network *network, double *rankmod)
 	return true;
 }
 
-void network_info_free(void *data)
-{
-	struct network_info *network = data;
-
-	l_free(network);
-}
-
-static struct network_info *network_info_get(const char *ssid,
-						enum security security)
-{
-	struct network_info *network;
-
-	network = known_networks_find(ssid, security);
-
-	if (!network) {
-		struct network_info search;
-
-		search.type = security;
-		strcpy(search.ssid, ssid);
-
-		network = l_queue_find(networks, network_info_match, &search);
-	}
-
-	if (!network) {
-		network = l_new(struct network_info, 1);
-		strcpy(network->ssid, ssid);
-		network->type = security;
-
-		l_queue_push_tail(networks, network);
-	}
-
-	network->seen_count++;
-
-	return network;
-}
-
-static void network_info_put(struct network_info *network)
-{
-	if (--network->seen_count)
-		return;
-
-	if (!networks || network->is_known)
-		return;
-
-	l_queue_remove(networks, network);
-	network_info_free(network);
-}
-
 struct network *network_create(struct station *station, const char *ssid,
 				enum security security)
 {
@@ -310,7 +220,12 @@ struct network *network_create(struct station *station, const char *ssid,
 
 	network = l_new(struct network, 1);
 	network->station = station;
-	network->info = network_info_get(ssid, security);
+	strcpy(network->ssid, ssid);
+	network->security = security;
+
+	network->info = known_networks_find(ssid, security);
+	if (network->info)
+		network->info->seen_count++;
 
 	network->bss_list = l_queue_new();
 	network->blacklist = l_queue_new();
@@ -320,7 +235,7 @@ struct network *network_create(struct station *station, const char *ssid,
 
 const char *network_get_ssid(const struct network *network)
 {
-	return network->info->ssid;
+	return network->ssid;
 }
 
 const char *network_get_path(const struct network *network)
@@ -330,11 +245,24 @@ const char *network_get_path(const struct network *network)
 
 enum security network_get_security(const struct network *network)
 {
-	return network->info->type;
+	return network->security;
 }
 
 const uint8_t *network_get_psk(struct network *network)
 {
+	if (network->psk)
+		return network->psk;
+
+	network->psk = l_malloc(32);
+
+	if (crypto_psk_from_passphrase(network->passphrase,
+					(unsigned char *)network->ssid,
+					strlen(network->ssid),
+					network->psk) < 0) {
+		l_free(network->psk);
+		network->psk = NULL;
+	}
+
 	return network->psk;
 }
 
@@ -343,6 +271,23 @@ const char *network_get_passphrase(const struct network *network)
 	return network->passphrase;
 }
 
+bool network_set_passphrase(struct network *network, const char *passphrase)
+{
+	if (network_get_security(network) != SECURITY_PSK)
+		return false;
+
+	if (!crypto_passphrase_is_valid(passphrase))
+		return false;
+
+	if (!network_settings_load(network))
+		network->settings = l_settings_new();
+
+	network_reset_passphrase(network);
+	network->passphrase = l_strdup(passphrase);
+
+	return true;
+}
+
 struct l_queue *network_get_secrets(const struct network *network)
 {
 	return network->secrets;
@@ -350,11 +295,11 @@ struct l_queue *network_get_secrets(const struct network *network)
 
 bool network_set_psk(struct network *network, const uint8_t *psk)
 {
-	if (network->info->type != SECURITY_PSK)
+	if (network_get_security(network) != SECURITY_PSK)
 		return false;
 
 	if (!network_settings_load(network))
-		return false;
+		network->settings = l_settings_new();
 
 	network_reset_psk(network);
 	network->psk = l_memdup(psk, 32);
@@ -422,12 +367,13 @@ static bool network_set_8021x_secrets(struct network *network)
 
 static int network_load_psk(struct network *network, bool need_passphrase)
 {
+	const char *ssid = network_get_ssid(network);
+	enum security security = network_get_security(network);
 	size_t len;
 	const char *psk = l_settings_get_value(network->settings,
 						"Security", "PreSharedKey");
 	char *passphrase = l_settings_get_string(network->settings,
 						"Security", "Passphrase");
-	struct network_info *info = network->info;
 	int r;
 
 	/* PSK can be generated from the passphrase but not the other way */
@@ -447,7 +393,7 @@ static int network_load_psk(struct network *network, bool need_passphrase)
 
 		network_reset_psk(network);
 
-		path = storage_get_network_file_path(info->type, info->ssid);
+		path = storage_get_network_file_path(security, ssid);
 		l_error("%s: invalid PreSharedKey format", path);
 		l_free(path);
 
@@ -456,8 +402,8 @@ static int network_load_psk(struct network *network, bool need_passphrase)
 	}
 
 	network->psk = l_malloc(32);
-	r = crypto_psk_from_passphrase(passphrase, (uint8_t *) info->ssid,
-					strlen(info->ssid), network->psk);
+	r = crypto_psk_from_passphrase(passphrase, (uint8_t *) ssid,
+					strlen(ssid), network->psk);
 	if (!r) {
 		network->update_psk = true;
 		return 0;
@@ -478,13 +424,14 @@ static int network_load_psk(struct network *network, bool need_passphrase)
 void network_sync_psk(struct network *network)
 {
 	struct l_settings *fs_settings;
+	const char *ssid = network_get_ssid(network);
 
 	if (!network->update_psk)
 		return;
 
 	network->update_psk = false;
 
-	fs_settings = storage_network_open(SECURITY_PSK, network->info->ssid);
+	fs_settings = storage_network_open(SECURITY_PSK, ssid);
 
 	if (network->psk) {
 		char *hex = l_util_hexstring(network->psk, 32);
@@ -510,12 +457,39 @@ void network_sync_psk(struct network *network)
 	}
 
 	if (fs_settings) {
-		storage_network_sync(SECURITY_PSK, network->info->ssid,
-					fs_settings);
+		storage_network_sync(SECURITY_PSK, ssid, fs_settings);
 		l_settings_free(fs_settings);
 	} else
-		storage_network_sync(SECURITY_PSK, network->info->ssid,
-					network->settings);
+		storage_network_sync(SECURITY_PSK, ssid, network->settings);
+}
+
+const struct network_info *network_get_info(const struct network *network)
+{
+	return network->info;
+}
+
+static void add_known_frequency(void *data, void *user_data)
+{
+	struct scan_bss *bss = data;
+	struct network_info *info = user_data;
+
+	known_network_add_frequency(info, bss->frequency);
+}
+
+void network_set_info(struct network *network, struct network_info *info)
+{
+	if (info) {
+		network->info = info;
+		network->info->seen_count++;
+
+		l_queue_foreach(network->bss_list, add_known_frequency, info);
+	} else {
+		network->info->seen_count--;
+		network->info = NULL;
+	}
+
+	l_dbus_property_changed(dbus_get_bus(), network_get_path(network),
+					IWD_NETWORK_INTERFACE, "KnownNetwork");
 }
 
 static inline bool __bss_is_sae(const struct scan_bss *bss,
@@ -629,10 +603,62 @@ void network_connect_failed(struct network *network)
 	l_queue_clear(network->blacklist, NULL);
 }
 
+static bool hotspot_info_matches(struct network *network,
+					const struct network_info *info)
+{
+	struct scan_bss *bss;
+
+	if (!network->is_hs20 || !info->is_hotspot)
+		return false;
+
+	bss = network_bss_select(network, true);
+
+	if (network_info_match_hessid(info, bss->hessid))
+		return true;
+
+	if (network_info_match_roaming_consortium(info, bss->rc_ie,
+							bss->rc_ie[1] + 2,
+							NULL))
+		return true;
+
+	return false;
+}
+
+static bool match_hotspot_network(const struct network_info *info,
+					void *user_data)
+{
+	struct network *network = user_data;
+
+	if (!hotspot_info_matches(network, info))
+		return false;
+
+	network_set_info(network, (struct network_info *) info);
+
+	return true;
+}
+
 bool network_bss_add(struct network *network, struct scan_bss *bss)
 {
-	return l_queue_insert(network->bss_list, bss,
-					scan_bss_rank_compare, NULL);
+	if (!l_queue_insert(network->bss_list, bss, scan_bss_rank_compare,
+									NULL))
+		return false;
+
+	if (network->info)
+		known_network_add_frequency(network->info, bss->frequency);
+
+	/* Done if BSS is not HS20 or we already have network_info set */
+	if (!bss->hs20_capable)
+		return true;
+	else
+		network->is_hs20 = true;
+
+	if (network->info)
+		return true;
+
+	/* Set the network_info to a matching hotspot entry, if found */
+	known_networks_foreach(match_hotspot_network, network);
+
+	return true;
 }
 
 bool network_bss_list_isempty(struct network *network)
@@ -724,6 +750,7 @@ static void passphrase_callback(enum agent_result result,
 {
 	struct network *network = user_data;
 	struct station *station = network->station;
+	const char *ssid = network_get_ssid(network);
 	struct scan_bss *bss;
 	int r;
 
@@ -754,8 +781,7 @@ static void passphrase_callback(enum agent_result result,
 	network_reset_psk(network);
 	network->psk = l_malloc(32);
 	r = crypto_psk_from_passphrase(passphrase,
-					(uint8_t *) network->info->ssid,
-					strlen(network->info->ssid),
+					(uint8_t *) ssid, strlen(ssid),
 					network->psk);
 	if (r) {
 		struct l_dbus_message *error;
@@ -807,9 +833,6 @@ static struct l_dbus_message *network_connect_psk(struct network *network,
 	 */
 	bool need_passphrase = bss_is_sae(bss);
 
-	l_debug("ask_passphrase: %s",
-		network->ask_passphrase ? "true" : "false");
-
 	if  (!network_settings_load(network)) {
 		network->settings = l_settings_new();
 		network->ask_passphrase = true;
@@ -817,6 +840,9 @@ static struct l_dbus_message *network_connect_psk(struct network *network,
 		network->ask_passphrase =
 			network_load_psk(network, need_passphrase) < 0;
 
+	l_debug("ask_passphrase: %s",
+		network->ask_passphrase ? "true" : "false");
+
 	if (network->ask_passphrase) {
 		network->ask_passphrase = false;
 
@@ -1090,8 +1116,11 @@ static struct l_dbus_message *network_connect(struct l_dbus *dbus,
 
 	l_debug("");
 
-	if (station_is_busy(station))
-		return dbus_error_busy(message);
+	if (network == station_get_connected_network(station))
+		/*
+		 * The requested network is already connected, return success.
+		 */
+		return l_dbus_message_new_method_return(message);
 
 	/*
 	 * Select the best BSS to use at this time.  If we have to query the
@@ -1134,7 +1163,6 @@ void network_connect_new_hidden_network(struct network *network,
 	 * network_sync_psk or network_connected will save this network
 	 * as hidden and trigger an update to the hidden networks count.
 	 */
-	network->info->is_hidden = true;
 
 	bss = network_bss_select(network, true);
 	if (!bss) {
@@ -1172,6 +1200,14 @@ void network_blacklist_add(struct network *network, struct scan_bss *bss)
 	l_queue_push_head(network->blacklist, bss);
 }
 
+const struct iovec *network_get_extra_ies(struct network *network,
+						size_t *num_elems)
+{
+	struct scan_bss *bss = network_bss_select(network, false);
+
+	return network_info_get_extra_ies(network->info, bss, num_elems);
+}
+
 static bool network_property_get_name(struct l_dbus *dbus,
 					struct l_dbus_message *message,
 					struct l_dbus_message_builder *builder,
@@ -1179,7 +1215,8 @@ static bool network_property_get_name(struct l_dbus *dbus,
 {
 	struct network *network = user_data;
 
-	l_dbus_message_builder_append_basic(builder, 's', network->info->ssid);
+	l_dbus_message_builder_append_basic(builder, 's',
+						network_get_ssid(network));
 	return true;
 }
 
@@ -1234,38 +1271,15 @@ static bool network_property_get_known_network(struct l_dbus *dbus,
 {
 	struct network *network = user_data;
 
-	if (!network->info->is_known)
+	if (!network->info)
 		return false;
 
 	l_dbus_message_builder_append_basic(builder, 'o',
-					known_network_get_path(network->info));
+					network_info_get_path(network->info));
 
 	return true;
 }
 
-static void setup_network_interface(struct l_dbus_interface *interface)
-{
-	l_dbus_interface_method(interface, "Connect", 0,
-				network_connect,
-				"", "");
-
-	l_dbus_interface_property(interface, "Name", 0, "s",
-					network_property_get_name, NULL);
-
-	l_dbus_interface_property(interface, "Connected", 0, "b",
-					network_property_is_connected,
-					NULL);
-
-	l_dbus_interface_property(interface, "Device", 0, "o",
-					network_property_get_device, NULL);
-
-	l_dbus_interface_property(interface, "Type", 0, "s",
-					network_property_get_type, NULL);
-
-	l_dbus_interface_property(interface, "KnownNetwork", 0, "o",
-				network_property_get_known_network, NULL);
-}
-
 bool network_register(struct network *network, const char *path)
 {
 	if (!l_dbus_object_add_interface(dbus_get_bus(), path,
@@ -1306,32 +1320,21 @@ void network_remove(struct network *network, int reason)
 	l_queue_destroy(network->secrets, eap_secret_info_free);
 	network->secrets = NULL;
 
-	l_queue_destroy(network->bss_list, NULL);
-	network_info_put(network->info);
+	if (network->info)
+		network->info->seen_count -= 1;
 
+	l_queue_destroy(network->bss_list, NULL);
 	l_queue_destroy(network->blacklist, NULL);
 
+	if (network->nai_realms)
+		l_strv_free(network->nai_realms);
+
+	if (network->rc_ie)
+		l_free(network->rc_ie);
+
 	l_free(network);
 }
 
-void network_init()
-{
-	if (!l_dbus_register_interface(dbus_get_bus(), IWD_NETWORK_INTERFACE,
-					setup_network_interface, NULL, false))
-		l_error("Unable to register %s interface",
-						IWD_NETWORK_INTERFACE);
-
-	networks = l_queue_new();
-}
-
-void network_exit()
-{
-	l_queue_destroy(networks, network_info_free);
-	networks = NULL;
-
-	l_dbus_unregister_interface(dbus_get_bus(), IWD_NETWORK_INTERFACE);
-}
-
 int network_rank_compare(const void *a, const void *b, void *user)
 {
 	const struct network *new_network = a;
@@ -1347,7 +1350,6 @@ void network_rank_update(struct network *network, bool connected)
 	 * here and in network_bss_select but those should be rare cases.
 	 */
 	struct scan_bss *best_bss = l_queue_peek_head(network->bss_list);
-	int rank;
 
 	/*
 	 * The rank should separate networks into four groups that use
@@ -1361,21 +1363,37 @@ void network_rank_update(struct network *network, bool connected)
 	 * Within the 2nd group the last connection time is the main factor,
 	 * for the other two groups it's the BSS rank - mainly signal strength.
 	 */
-	if (connected)
-		rank = INT_MAX;
-	else if (network->info->connected_time.tv_sec != 0) {
-		int n = network_find_rank_index(network->info);
+	if (connected) {
+		network->rank = INT_MAX;
+		return;
+	}
+
+	if (!network->info) { /* Not known, assign negative rank */
+		network->rank = (int) best_bss->rank - USHRT_MAX;
+		return;
+	}
+
+	if (network->info->connected_time != 0) {
+		int n = known_network_offset(network->info);
+
+		L_WARN_ON(n < 0);
 
 		if (n >= (int) L_ARRAY_SIZE(rankmod_table))
 			n = L_ARRAY_SIZE(rankmod_table) - 1;
 
-		rank = rankmod_table[n] * best_bss->rank + USHRT_MAX;
-	} else if (network->info->is_known)
-		rank = best_bss->rank;
-	else
-		rank = (int) best_bss->rank - USHRT_MAX; /* Negative rank */
+		network->rank = rankmod_table[n] * best_bss->rank + USHRT_MAX;
+	} else
+		network->rank = best_bss->rank;
+}
 
-	network->rank = rank;
+static void network_unset_hotspot(struct network *network, void *user_data)
+{
+	struct network_info *info = user_data;
+
+	if (network->info != info)
+		return;
+
+	network_set_info(network, NULL);
 }
 
 static void emit_known_network_changed(struct station *station, void *user_data)
@@ -1383,39 +1401,42 @@ static void emit_known_network_changed(struct station *station, void *user_data)
 	struct network_info *info = user_data;
 	struct network *network;
 
-	network = station_network_find(station, info->ssid, info->type);
-	if (!network)
+	if (!info->is_hotspot) {
+		network = station_network_find(station, info->ssid, info->type);
+		if (!network)
+			return;
+
+		network_set_info(network, NULL);
 		return;
-
-	l_dbus_property_changed(dbus_get_bus(),
-				network_get_path(network),
-				IWD_NETWORK_INTERFACE,
-				"KnownNetwork");
-}
-
-struct network_info *network_info_add_known(const char *ssid,
-						enum security security)
-{
-	struct network_info *network;
-	struct network_info search;
-
-	strcpy(search.ssid, ssid);
-	search.type = security;
-
-	network = l_queue_remove_if(networks, network_info_match, &search);
-	if (network) {
-		/* Promote network to is_known */
-		network->is_known = true;
-		station_foreach(emit_known_network_changed, network);
-		return network;
 	}
 
-	network = l_new(struct network_info, 1);
-	strcpy(network->ssid, ssid);
-	network->type = security;
-	network->is_known = true;
+	/* This is a removed hotspot */
+	station_network_foreach(station, network_unset_hotspot, info);
+}
 
-	return network;
+static void network_update_hotspot(struct network *network, void *user_data)
+{
+	struct network_info *info = user_data;
+
+	match_hotspot_network(info, network);
+}
+
+static void match_known_network(struct station *station, void *user_data)
+{
+	struct network_info *info = user_data;
+	struct network *network;
+
+	if (!info->is_hotspot) {
+		network = station_network_find(station, info->ssid, info->type);
+		if (!network)
+			return;
+
+		network_set_info(network, info);
+		return;
+	}
+
+	/* This is a new hotspot network */
+	station_network_foreach(station, network_update_hotspot, info);
 }
 
 static void disconnect_no_longer_known(struct station *station, void *user_data)
@@ -1429,20 +1450,67 @@ static void disconnect_no_longer_known(struct station *station, void *user_data)
 		station_disconnect(station);
 }
 
-void network_info_forget_known(struct network_info *network)
+static void known_networks_changed(enum known_networks_event event,
+					const struct network_info *info,
+					void *user_data)
 {
-	network->is_known = false;
-	memset(&network->connected_time, 0, sizeof(network->connected_time));
+	switch (event) {
+	case KNOWN_NETWORKS_EVENT_ADDED:
+		station_foreach(match_known_network, (void *) info);
 
-	station_foreach(emit_known_network_changed, network);
-	station_foreach(disconnect_no_longer_known, network);
-
-	/*
-	 * Network is no longer a Known Network, see if we still need to
-	 * keep the network_info around.
-	 */
-	if (network->seen_count)
-		l_queue_push_tail(networks, network);
-	else
-		network_info_free(network);
+		/* Syncs frequencies of newly known network */
+		known_network_frequency_sync((struct network_info *)info);
+		break;
+	case KNOWN_NETWORKS_EVENT_REMOVED:
+		station_foreach(disconnect_no_longer_known, (void *) info);
+		station_foreach(emit_known_network_changed, (void *) info);
+		break;
+	}
 }
+
+static void setup_network_interface(struct l_dbus_interface *interface)
+{
+	l_dbus_interface_method(interface, "Connect", 0,
+				network_connect,
+				"", "");
+
+	l_dbus_interface_property(interface, "Name", 0, "s",
+					network_property_get_name, NULL);
+
+	l_dbus_interface_property(interface, "Connected", 0, "b",
+					network_property_is_connected,
+					NULL);
+
+	l_dbus_interface_property(interface, "Device", 0, "o",
+					network_property_get_device, NULL);
+
+	l_dbus_interface_property(interface, "Type", 0, "s",
+					network_property_get_type, NULL);
+
+	l_dbus_interface_property(interface, "KnownNetwork", 0, "o",
+				network_property_get_known_network, NULL);
+}
+
+static int network_init(void)
+{
+	if (!l_dbus_register_interface(dbus_get_bus(), IWD_NETWORK_INTERFACE,
+					setup_network_interface, NULL, false))
+		l_error("Unable to register %s interface",
+						IWD_NETWORK_INTERFACE);
+
+	known_networks_watch =
+		known_networks_watch_add(known_networks_changed, NULL, NULL);
+
+	return 0;
+}
+
+static void network_exit(void)
+{
+	known_networks_watch_remove(known_networks_watch);
+	known_networks_watch = 0;
+
+	l_dbus_unregister_interface(dbus_get_bus(), IWD_NETWORK_INTERFACE);
+}
+
+IWD_MODULE(network, network_init, network_exit)
+IWD_MODULE_DEPENDS(network, known_networks)
