@@ -2,7 +2,7 @@
  *
  *  Wireless daemon for Linux
  *
- *  Copyright (C) 2013-2014  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2013-2019  Intel Corporation. All rights reserved.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -24,11 +24,14 @@
 #include <config.h>
 #endif
 
+#define _GNU_SOURCE
 #include <stdlib.h>
 #include <stdio.h>
 #include <errno.h>
 #include <linux/if_ether.h>
 #include <fnmatch.h>
+#include <unistd.h>
+#include <string.h>
 
 #include <ell/ell.h>
 
@@ -44,29 +47,44 @@
 #include "src/wiphy.h"
 #include "src/storage.h"
 #include "src/util.h"
+#include "src/common.h"
+#include "src/watchlist.h"
+#include "src/nl80211util.h"
+
+#define EXT_CAP_LEN 10
 
 static struct l_genl_family *nl80211 = NULL;
 static struct l_hwdb *hwdb;
 static char **whitelist_filter;
 static char **blacklist_filter;
+static int mac_randomize_bytes = 6;
 
 struct wiphy {
 	uint32_t id;
 	char name[20];
+	uint8_t permanent_addr[ETH_ALEN];
 	uint32_t feature_flags;
 	uint8_t ext_features[(NUM_NL80211_EXT_FEATURES + 7) / 8];
 	uint8_t max_num_ssids_per_scan;
-	bool support_scheduled_scan:1;
-	bool support_rekey_offload:1;
-	bool support_adhoc_rsn:1;
+	uint32_t max_roc_duration;
 	uint16_t supported_iftypes;
 	uint16_t supported_ciphers;
 	struct scan_freq_set *supported_freqs;
 	char *model_str;
 	char *vendor_str;
+	char *driver_str;
+	struct watchlist state_watches;
+	uint8_t extended_capabilities[EXT_CAP_LEN + 2]; /* max bitmap size + IE header */
+	uint8_t *iftype_extended_capabilities[NUM_NL80211_IFTYPES];
+	uint8_t rm_enabled_capabilities[7]; /* 5 size max + header */
 
+	bool support_scheduled_scan:1;
+	bool support_rekey_offload:1;
+	bool support_adhoc_rsn:1;
+	bool support_qos_set_map:1;
 	bool soft_rfkill : 1;
 	bool hard_rfkill : 1;
+	bool offchannel_tx_ok : 1;
 };
 
 static struct l_queue *wiphy_list = NULL;
@@ -90,15 +108,115 @@ enum ie_rsn_cipher_suite wiphy_select_cipher(struct wiphy *wiphy, uint16_t mask)
 
 	return 0;
 }
+
+enum ie_rsn_akm_suite wiphy_select_akm(struct wiphy *wiphy,
+					struct scan_bss *bss,
+					bool fils_capable_hint)
+{
+	struct ie_rsn_info info;
+	enum security security;
+
+	memset(&info, 0, sizeof(info));
+	scan_bss_get_rsn_info(bss, &info);
+
+	security = security_determine(bss->capability, &info);
+
+	/*
+	 * If FT is available, use FT authentication to keep the door open
+	 * for fast transitions.  Otherwise use SHA256 version if present.
+	 */
+	if (security == SECURITY_8021X) {
+		if (wiphy_has_feature(wiphy, NL80211_EXT_FEATURE_FILS_STA) &&
+				fils_capable_hint) {
+			if ((info.akm_suites &
+					IE_RSN_AKM_SUITE_FT_OVER_FILS_SHA384) &&
+					bss->rsne && bss->mde_present)
+				return IE_RSN_AKM_SUITE_FT_OVER_FILS_SHA384;
+
+			if ((info.akm_suites &
+					IE_RSN_AKM_SUITE_FT_OVER_FILS_SHA256) &&
+					bss->rsne && bss->mde_present)
+				return IE_RSN_AKM_SUITE_FT_OVER_FILS_SHA256;
+
+			if (info.akm_suites & IE_RSN_AKM_SUITE_FILS_SHA384)
+				return IE_RSN_AKM_SUITE_FILS_SHA384;
+
+			if (info.akm_suites & IE_RSN_AKM_SUITE_FILS_SHA256)
+				return IE_RSN_AKM_SUITE_FILS_SHA256;
+		}
+
+		if ((info.akm_suites & IE_RSN_AKM_SUITE_FT_OVER_8021X) &&
+				bss->rsne && bss->mde_present)
+			return IE_RSN_AKM_SUITE_FT_OVER_8021X;
+
+		if (info.akm_suites & IE_RSN_AKM_SUITE_8021X_SHA256)
+			return IE_RSN_AKM_SUITE_8021X_SHA256;
+
+		if (info.akm_suites & IE_RSN_AKM_SUITE_8021X)
+			return IE_RSN_AKM_SUITE_8021X;
+	} else if (security == SECURITY_PSK) {
+		/*
+		 * Prefer connecting to SAE/WPA3 network, but only if SAE is
+		 * supported, we are MFP capable, and the AP has set the MFPR
+		 * bit. If any of these conditions are not met, we can fallback
+		 * to WPA2 (if the AKM is present).
+		 */
+		if (wiphy->supported_ciphers & IE_RSN_CIPHER_SUITE_BIP &&
+				wiphy_has_feature(wiphy, NL80211_FEATURE_SAE) &&
+				info.mfpr) {
+			if (info.akm_suites &
+					IE_RSN_AKM_SUITE_FT_OVER_SAE_SHA256)
+				return IE_RSN_AKM_SUITE_FT_OVER_SAE_SHA256;
+
+			if (info.akm_suites & IE_RSN_AKM_SUITE_SAE_SHA256)
+				return IE_RSN_AKM_SUITE_SAE_SHA256;
+		}
+
+		if ((info.akm_suites & IE_RSN_AKM_SUITE_FT_USING_PSK) &&
+				bss->rsne && bss->mde_present)
+			return IE_RSN_AKM_SUITE_FT_USING_PSK;
+
+		if (info.akm_suites & IE_RSN_AKM_SUITE_PSK_SHA256)
+			return IE_RSN_AKM_SUITE_PSK_SHA256;
+
+		if (info.akm_suites & IE_RSN_AKM_SUITE_PSK)
+			return IE_RSN_AKM_SUITE_PSK;
+	} else if (security == SECURITY_NONE) {
+		if (info.akm_suites & IE_RSN_AKM_SUITE_OWE)
+			return IE_RSN_AKM_SUITE_OWE;
+	}
+
+	return 0;
+}
+
+static struct wiphy *wiphy_new(uint32_t id)
+{
+	struct wiphy *wiphy = l_new(struct wiphy, 1);
+
+	wiphy->id = id;
+	wiphy->supported_freqs = scan_freq_set_new();
+	watchlist_init(&wiphy->state_watches, NULL);
+	wiphy->extended_capabilities[0] = IE_TYPE_EXTENDED_CAPABILITIES;
+	wiphy->extended_capabilities[1] = EXT_CAP_LEN;
+
+	return wiphy;
+}
+
 static void wiphy_free(void *data)
 {
 	struct wiphy *wiphy = data;
+	uint32_t i;
 
 	l_debug("Freeing wiphy %s[%u]", wiphy->name, wiphy->id);
 
+	for (i = 0; i < NUM_NL80211_IFTYPES; i++)
+		l_free(wiphy->iftype_extended_capabilities[i]);
+
 	scan_freq_set_free(wiphy->supported_freqs);
+	watchlist_destroy(&wiphy->state_watches);
 	l_free(wiphy->model_str);
 	l_free(wiphy->vendor_str);
+	l_free(wiphy->driver_str);
 	l_free(wiphy);
 }
 
@@ -149,9 +267,12 @@ check_blacklist:
 
 const char *wiphy_get_path(struct wiphy *wiphy)
 {
-	static char path[15];
+	static char path[256];
+
+	L_WARN_ON(snprintf(path, sizeof(path), "%s/%d", IWD_BASE_PATH,
+				wiphy->id) >= (int) sizeof(path));
+	path[sizeof(path) - 1] = '\0';
 
-	snprintf(path, sizeof(path), "/%d", wiphy->id);
 	return path;
 }
 
@@ -163,6 +284,12 @@ uint32_t wiphy_get_supported_bands(struct wiphy *wiphy)
 	return scan_freq_set_get_bands(wiphy->supported_freqs);
 }
 
+const struct scan_freq_set *wiphy_get_supported_freqs(
+						const struct wiphy *wiphy)
+{
+	return wiphy->supported_freqs;
+}
+
 bool wiphy_can_connect(struct wiphy *wiphy, struct scan_bss *bss)
 {
 	struct ie_rsn_info rsn_info;
@@ -181,6 +308,17 @@ bool wiphy_can_connect(struct wiphy *wiphy, struct scan_bss *bss)
 		if (rsn_info.mfpr && !wiphy_select_cipher(wiphy,
 					rsn_info.group_management_cipher))
 			return false;
+
+		/*
+		 * if the AP ONLY supports SAE/WPA3, then we can only connect
+		 * if the wiphy feature is supported. Otherwise the AP may list
+		 * SAE as one of the AKM's but also support PSK (hybrid). In
+		 * this case we still want to allow a connection even if SAE
+		 * is not supported.
+		 */
+		if (IE_AKM_IS_SAE(rsn_info.akm_suites) &&
+				!wiphy_has_feature(wiphy, NL80211_FEATURE_SAE))
+			return false;
 	} else if (r != -ENOENT)
 		return false;
 
@@ -192,6 +330,24 @@ bool wiphy_has_feature(struct wiphy *wiphy, uint32_t feature)
 	return wiphy->feature_flags & feature;
 }
 
+bool wiphy_can_randomize_mac_addr(struct wiphy *wiphy)
+{
+	return wiphy_has_feature(wiphy, NL80211_FEATURE_SCAN_RANDOM_MAC_ADDR);
+}
+
+bool wiphy_rrm_capable(struct wiphy *wiphy)
+{
+	if (wiphy_has_feature(wiphy,
+				NL80211_FEATURE_DS_PARAM_SET_IE_IN_PROBES) &&
+			wiphy_has_feature(wiphy, NL80211_FEATURE_QUIET))
+		return true;
+
+	if (wiphy_has_ext_feature(wiphy, NL80211_EXT_FEATURE_RRM))
+		return true;
+
+	return false;
+}
+
 bool wiphy_has_ext_feature(struct wiphy *wiphy, uint32_t feature)
 {
 	return feature < sizeof(wiphy->ext_features) * 8 &&
@@ -203,23 +359,115 @@ uint8_t wiphy_get_max_num_ssids_per_scan(struct wiphy *wiphy)
 	return wiphy->max_num_ssids_per_scan;
 }
 
+uint32_t wiphy_get_max_roc_duration(struct wiphy *wiphy)
+{
+	return wiphy->max_roc_duration;
+}
+
 bool wiphy_supports_adhoc_rsn(struct wiphy *wiphy)
 {
 	return wiphy->support_adhoc_rsn;
 }
 
-static char **wiphy_get_supported_iftypes(struct wiphy *wiphy)
+bool wiphy_can_offchannel_tx(struct wiphy *wiphy)
 {
-	char **ret = l_new(char *,
-			__builtin_popcount(wiphy->supported_iftypes) + 1);
+	return wiphy->offchannel_tx_ok;
+}
+
+bool wiphy_supports_qos_set_map(struct wiphy *wiphy)
+{
+	return wiphy->support_qos_set_map;
+}
+
+const char *wiphy_get_driver(struct wiphy *wiphy)
+{
+	return wiphy->driver_str;
+}
+
+const char *wiphy_get_name(struct wiphy *wiphy)
+{
+	return wiphy->name;
+}
+
+const uint8_t *wiphy_get_permanent_address(struct wiphy *wiphy)
+{
+	return wiphy->permanent_addr;
+}
+
+const uint8_t *wiphy_get_extended_capabilities(struct wiphy *wiphy,
+							uint32_t iftype)
+{
+	if (wiphy->iftype_extended_capabilities[iftype])
+		return wiphy->iftype_extended_capabilities[iftype];
+
+	return wiphy->extended_capabilities;
+}
+
+const uint8_t *wiphy_get_rm_enabled_capabilities(struct wiphy *wiphy)
+{
+	if (!wiphy_rrm_capable(wiphy))
+		return NULL;
+
+	return wiphy->rm_enabled_capabilities;
+}
+
+void wiphy_generate_random_address(struct wiphy *wiphy, uint8_t addr[static 6])
+{
+	switch (mac_randomize_bytes) {
+	case 6:
+		l_getrandom(addr, 6);
+
+		/* Set the locally administered bit */
+		addr[0] |= 0x2;
+
+		/* Reset multicast bit */
+		addr[0] &= 0xfe;
+		break;
+	case 3:
+		l_getrandom(addr + 3, 3);
+		memcpy(addr, wiphy->permanent_addr, 3);
+		break;
+	}
+
+	/*
+	 * Constrain the last NIC byte to 0x00 .. 0xfe, otherwise we might be
+	 * able to generate an address of 0xff 0xff 0xff which might be
+	 * interpreted as a vendor broadcast.  Similarly, 0x00 0x00 0x00 is
+	 * also not valid
+	 */
+	addr[5] &= 0xfe;
+	if (util_mem_is_zero(addr + 3, 3))
+		addr[5] = 0x01;
+}
+
+bool wiphy_constrain_freq_set(const struct wiphy *wiphy,
+						struct scan_freq_set *set)
+{
+	scan_freq_set_constrain(set, wiphy->supported_freqs);
+
+	if (!scan_freq_set_get_bands(set))
+		/* The set is empty. */
+		return false;
+
+	return true;
+}
+
+static char **wiphy_get_supported_iftypes(struct wiphy *wiphy, uint16_t mask)
+{
+	uint16_t supported_mask = wiphy->supported_iftypes & mask;
+	char **ret = l_new(char *, __builtin_popcount(supported_mask) + 1);
 	unsigned int i;
 	unsigned int j;
 
-	for (j = 0, i = 0; i < sizeof(wiphy->supported_iftypes) * 8; i++) {
-		if (!(wiphy->supported_iftypes & (1 << i)))
+	for (j = 0, i = 0; i < sizeof(supported_mask) * 8; i++) {
+		const char *str;
+
+		if (!(supported_mask & (1 << i)))
 			continue;
 
-		ret[j++] = l_strdup(dbus_iftype_to_string(i + 1));
+		str = dbus_iftype_to_string(i + 1);
+		if (str)
+			ret[j++] = l_strdup(str);
 	}
 
 	return ret;
@@ -233,12 +481,25 @@ bool wiphy_supports_iftype(struct wiphy *wiphy, uint32_t iftype)
 	return wiphy->supported_iftypes & (1 << (iftype - 1));
 }
 
+uint32_t wiphy_state_watch_add(struct wiphy *wiphy,
+				wiphy_state_watch_func_t func,
+				void *user_data, wiphy_destroy_func_t destroy)
+{
+	return watchlist_add(&wiphy->state_watches, func, user_data, destroy);
+}
+
+bool wiphy_state_watch_remove(struct wiphy *wiphy, uint32_t id)
+{
+	return watchlist_remove(&wiphy->state_watches, id);
+}
+
 static void wiphy_print_basic_info(struct wiphy *wiphy)
 {
 	uint32_t bands;
 	char buf[1024];
 
 	l_info("Wiphy: %d, Name: %s", wiphy->id, wiphy->name);
+	l_info("\tPermanent Address: "MAC, MAC_STR(wiphy->permanent_addr));
 
 	bands = scan_freq_set_get_bands(wiphy->supported_freqs);
 
@@ -274,7 +535,7 @@ static void wiphy_print_basic_info(struct wiphy *wiphy)
 	}
 
 	if (wiphy->supported_iftypes) {
-		char **iftypes = wiphy_get_supported_iftypes(wiphy);
+		char **iftypes = wiphy_get_supported_iftypes(wiphy, ~0);
 		char *joined = l_strjoinv(iftypes, ' ');
 
 		l_info("\tSupported iftypes: %s", joined);
@@ -299,6 +560,10 @@ static void parse_supported_commands(struct wiphy *wiphy,
 			break;
 		case NL80211_CMD_SET_REKEY_OFFLOAD:
 			wiphy->support_rekey_offload = true;
+			break;
+		case NL80211_CMD_SET_QOS_MAP:
+			wiphy->support_qos_set_map = true;
+			break;
 		}
 	}
 }
@@ -408,14 +673,57 @@ static void parse_supported_iftypes(struct wiphy *wiphy,
 	}
 }
 
-static void wiphy_parse_attributes(struct wiphy *wiphy,
-					struct l_genl_attr *attr)
+static void parse_iftype_extended_capabilities(struct wiphy *wiphy,
+						struct l_genl_attr *attr)
 {
+	uint16_t type;
+	uint16_t len;
+	const void *data;
+	struct l_genl_attr nested;
+
+	while (l_genl_attr_next(attr, &type, &len, &data)) {
+		uint32_t iftype;
+
+		if (!l_genl_attr_recurse(attr, &nested))
+			continue;
+
+		if (!l_genl_attr_next(&nested, &type, &len, &data))
+			continue;
+
+		if (type != NL80211_ATTR_IFTYPE)
+			continue;
+
+		iftype = l_get_u32(data);
+
+		if (!l_genl_attr_next(&nested, &type, &len, &data))
+			continue;
+
+		if (type != NL80211_ATTR_EXT_CAPA)
+			continue;
+
+		wiphy->iftype_extended_capabilities[iftype] =
+					l_new(uint8_t, EXT_CAP_LEN + 2);
+		wiphy->iftype_extended_capabilities[iftype][0] =
+					IE_TYPE_EXTENDED_CAPABILITIES;
+		wiphy->iftype_extended_capabilities[iftype][1] =
+					EXT_CAP_LEN;
+		memcpy(wiphy->iftype_extended_capabilities[iftype] + 2,
+				data, minsize(len, EXT_CAP_LEN));
+	}
+}
+
+static void wiphy_parse_attributes(struct wiphy *wiphy,
+					struct l_genl_msg *msg)
+{
+	struct l_genl_attr attr;
 	struct l_genl_attr nested;
 	uint16_t type, len;
 	const void *data;
 
-	while (l_genl_attr_next(attr, &type, &len, &data)) {
+	if (!l_genl_attr_init(&attr, msg))
+		return;
+
+	while (l_genl_attr_next(&attr, &type, &len, &data)) {
 		switch (type) {
 		case NL80211_ATTR_FEATURE_FLAGS:
 			if (len != sizeof(uint32_t))
@@ -431,7 +739,7 @@ static void wiphy_parse_attributes(struct wiphy *wiphy,
 			memcpy(wiphy->ext_features, data, len);
 			break;
 		case NL80211_ATTR_SUPPORTED_COMMANDS:
-			if (l_genl_attr_recurse(attr, &nested))
+			if (l_genl_attr_recurse(&attr, &nested))
 				parse_supported_commands(wiphy, &nested);
 
 			break;
@@ -439,7 +747,7 @@ static void wiphy_parse_attributes(struct wiphy *wiphy,
 			parse_supported_ciphers(wiphy, data, len);
 			break;
 		case NL80211_ATTR_WIPHY_BANDS:
-			if (l_genl_attr_recurse(attr, &nested))
+			if (l_genl_attr_recurse(&attr, &nested))
 				parse_supported_bands(wiphy, &nested);
 
 			break;
@@ -454,113 +762,73 @@ static void wiphy_parse_attributes(struct wiphy *wiphy,
 			wiphy->support_adhoc_rsn = true;
 			break;
 		case NL80211_ATTR_SUPPORTED_IFTYPES:
-			if (l_genl_attr_recurse(attr, &nested))
+			if (l_genl_attr_recurse(&attr, &nested))
 				parse_supported_iftypes(wiphy, &nested);
 			break;
+		case NL80211_ATTR_OFFCHANNEL_TX_OK:
+			wiphy->offchannel_tx_ok = true;
+			break;
+		case NL80211_ATTR_EXT_CAPA:
+			memcpy(wiphy->extended_capabilities + 2,
+				data, minsize(EXT_CAP_LEN, len));
+			break;
+		case NL80211_ATTR_IFTYPE_EXT_CAPA:
+			if (!l_genl_attr_recurse(&attr, &nested))
+				break;
+
+			parse_iftype_extended_capabilities(wiphy, &nested);
+			break;
+		case NL80211_ATTR_MAX_REMAIN_ON_CHANNEL_DURATION:
+			if (len != 4)
+				l_warn("Invalid MAX_ROC_DURATION attribute");
+			else
+				wiphy->max_roc_duration = *((uint32_t *) data);
+			break;
 		}
 	}
-
 }
 
-static bool wiphy_parse_id_and_name(struct l_genl_attr *attr, uint32_t *out_id,
-					const char **out_name,
-					uint32_t *out_name_len)
+static bool wiphy_get_driver_name(struct wiphy *wiphy)
 {
-	uint16_t type, len;
-	const void *data;
-	uint32_t id;
-	const char *name;
-	uint32_t name_len;
+	L_AUTO_FREE_VAR(char *, driver_link) = NULL;
+	char driver_path[256];
+	ssize_t len;
 
-	/*
-	 * The wiphy attribute, name and generation are always the first
-	 * three attributes (in that order) in every NEW_WIPHY & DEL_WIPHY
-	 * message.  If not, then error out with a warning and ignore the
-	 * whole message.
-	 */
-	if (!l_genl_attr_next(attr, &type, &len, &data))
+	driver_link = l_strdup_printf("/sys/class/ieee80211/%s/device/driver",
+					wiphy->name);
+	len = readlink(driver_link, driver_path, sizeof(driver_path) - 1);
+
+	if (len == -1) {
+		l_error("Can't read %s: %s", driver_link, strerror(errno));
 		return false;
+	}
 
-	if (type != NL80211_ATTR_WIPHY)
-		return false;
-
-	if (len != sizeof(uint32_t))
-		return false;
-
-	id = *((uint32_t *) data);
-
-	if (!l_genl_attr_next(attr, &type, &len, &data))
-		return false;
-
-	if (type != NL80211_ATTR_WIPHY_NAME)
-		return false;
-
-	if (len > sizeof(((struct wiphy *) 0)->name))
-		return false;
-
-	name = data;
-	name_len = len;
-
-	if (!l_genl_attr_next(attr, &type, &len, &data))
-		return false;
-
-	if (type != NL80211_ATTR_GENERATION)
-		return false;
-
-	if (len != sizeof(uint32_t))
-		return false;
-
-	/*
-	 * TODO: Handle GENERATION.  In theory if we detect a changed generation
-	 * number during a dump, it means that our dump needs to be re-started
-	 */
-
-	if (out_id)
-		*out_id = id;
-
-	if (out_name)
-		*out_name = name;
-
-	if (out_name_len)
-		*out_name_len = name_len;
-
+	driver_path[len] = '\0';
+	wiphy->driver_str = l_strdup(basename(driver_path));
 	return true;
 }
 
-static void wiphy_dump_callback(struct l_genl_msg *msg, void *user_data)
+static int wiphy_get_permanent_addr_from_sysfs(struct wiphy *wiphy)
 {
-	struct wiphy *wiphy;
-	struct l_genl_attr attr;
-	uint32_t id;
-	const char *name;
-	uint32_t name_len;
+	char addr[32];
+	ssize_t len;
 
-	l_debug("");
-
-	if (!l_genl_attr_init(&attr, msg))
-		return;
-
-	/*
-	 * In most cases multiple of these message will be sent
-	 * since the information included can not fit into a single
-	 * message.
-	 */
-	if (!wiphy_parse_id_and_name(&attr, &id, &name, &name_len))
-		return;
-
-	wiphy = l_queue_find(wiphy_list, wiphy_match, L_UINT_TO_PTR(id));
-	if (!wiphy) {
-		if (!wiphy_is_managed(name))
-			return;
-
-		wiphy = l_new(struct wiphy, 1);
-		wiphy->id = id;
-		wiphy->supported_freqs = scan_freq_set_new();
-		l_queue_push_head(wiphy_list, wiphy);
+	len = read_file(addr, sizeof(addr),
+				"/sys/class/ieee80211/%s/macaddress",
+				wiphy->name);
+	if (len != 18) {
+		if (len < 0)
+			return -errno;
+		return -EINVAL;
 	}
 
-	memcpy(wiphy->name, name, name_len);
-	wiphy_parse_attributes(wiphy, &attr);
+	/* Sysfs appends a \n at the end, strip it */
+	addr[17] = '\0';
+
+	if (!util_string_to_address(addr, wiphy->permanent_addr))
+		return -EINVAL;
+
+	return 0;
 }
 
 static void wiphy_register(struct wiphy *wiphy)
@@ -603,6 +871,8 @@ static void wiphy_register(struct wiphy *wiphy)
 		l_hwdb_lookup_free(entries);
 	}
 
+	wiphy_get_driver_name(wiphy);
+
 	if (!l_dbus_object_add_interface(dbus, wiphy_get_path(wiphy),
 					IWD_WIPHY_INTERFACE, wiphy))
 		l_info("Unable to add the %s interface to %s",
@@ -615,175 +885,116 @@ static void wiphy_register(struct wiphy *wiphy)
 				wiphy_get_path(wiphy));
 }
 
-static void wiphy_dump_done(void *user)
-{
-	const struct l_queue_entry *wiphy_entry;
-
-	for (wiphy_entry = l_queue_get_entries(wiphy_list); wiphy_entry;
-					wiphy_entry = wiphy_entry->next) {
-		struct wiphy *wiphy = wiphy_entry->data;
-
-		wiphy_register(wiphy);
-
-		wiphy_print_basic_info(wiphy);
-	}
-}
-
-static void wiphy_new_wiphy_event(struct l_genl_msg *msg)
+struct wiphy *wiphy_create(uint32_t wiphy_id, const char *name)
 {
 	struct wiphy *wiphy;
-	struct l_genl_attr attr;
-	uint32_t id;
-	const char *name;
-	uint32_t name_len;
-
-	l_debug("");
-
-	if (!l_genl_attr_init(&attr, msg))
-		return;
-
-	if (!wiphy_parse_id_and_name(&attr, &id, &name, &name_len))
-		return;
-
-	wiphy = l_queue_find(wiphy_list, wiphy_match, L_UINT_TO_PTR(id));
-	if (wiphy) {
-		/*
-		 * WIPHY_NAME is a NLA_NUL_STRING, so the kernel
-		 * enforces the data to be null terminated
-		 */
-		if (strcmp(wiphy->name, name)) {
-			struct l_dbus *dbus = dbus_get_bus();
-
-			memcpy(wiphy->name, name, name_len);
-			l_dbus_property_changed(dbus, wiphy_get_path(wiphy),
-						IWD_WIPHY_INTERFACE, "Name");
-		}
-
-		return;
-	}
 
 	if (!wiphy_is_managed(name))
-		return;
+		return NULL;
 
-	wiphy = l_new(struct wiphy, 1);
-	wiphy->id = id;
-	memcpy(wiphy->name, name, name_len);
-	wiphy->supported_freqs = scan_freq_set_new();
+	wiphy = wiphy_new(wiphy_id);
+	l_strlcpy(wiphy->name, name, sizeof(wiphy->name));
 	l_queue_push_head(wiphy_list, wiphy);
 
-	wiphy_parse_attributes(wiphy, &attr);
-	wiphy_print_basic_info(wiphy);
-
 	wiphy_register(wiphy);
+	return wiphy;
 }
 
-static void wiphy_del_wiphy_event(struct l_genl_msg *msg)
+void wiphy_update_from_genl(struct wiphy *wiphy, const char *name,
+				struct l_genl_msg *msg)
 {
-	struct wiphy *wiphy;
-	struct l_genl_attr attr;
-	uint32_t id;
-
 	l_debug("");
 
-	if (!l_genl_attr_init(&attr, msg))
+	if (strncmp(wiphy->name, name, sizeof(wiphy->name))) {
+		struct l_dbus *dbus = dbus_get_bus();
+
+		l_strlcpy(wiphy->name, name, sizeof(wiphy->name));
+		l_dbus_property_changed(dbus, wiphy_get_path(wiphy),
+					IWD_WIPHY_INTERFACE, "Name");
+	}
+
+	wiphy_parse_attributes(wiphy, msg);
+}
+
+static void wiphy_set_station_capability_bits(struct wiphy *wiphy)
+{
+	uint8_t *ext_capa;
+	bool anqp_disabled;
+
+	/* No per-type capabilities exist for station, just copy the global */
+	if (!wiphy->iftype_extended_capabilities[NL80211_IFTYPE_STATION]) {
+		wiphy->iftype_extended_capabilities[NL80211_IFTYPE_STATION] =
+					l_new(uint8_t, EXT_CAP_LEN + 2);
+
+		memcpy(wiphy->iftype_extended_capabilities[
+						NL80211_IFTYPE_STATION],
+						wiphy->extended_capabilities,
+						EXT_CAP_LEN + 2);
+	}
+
+	ext_capa = wiphy->iftype_extended_capabilities[NL80211_IFTYPE_STATION];
+
+	if (!l_settings_get_bool(iwd_get_config(), "General", "DisableANQP",
+				&anqp_disabled))
+		anqp_disabled = true;
+
+	/* Set BSS Transition Management */
+	util_set_bit(ext_capa + 2, 19);
+
+	/* Set Interworking */
+	if (!anqp_disabled)
+		util_set_bit(ext_capa + 2, 31);
+
+	/* Set QoS Map */
+	if (wiphy->support_qos_set_map)
+		util_set_bit(ext_capa + 2, 32);
+
+	/* Set FILS */
+	util_set_bit(ext_capa + 2, 72);
+}
+
+static void wiphy_setup_rm_enabled_capabilities(struct wiphy *wiphy)
+{
+	/* Nothing to do */
+	if (!wiphy_rrm_capable(wiphy))
 		return;
 
-	if (!wiphy_parse_id_and_name(&attr, &id, NULL, NULL))
-		return;
+	wiphy->rm_enabled_capabilities[0] = IE_TYPE_RM_ENABLED_CAPABILITIES;
+	wiphy->rm_enabled_capabilities[1] = 5;
 
-	wiphy = l_queue_remove_if(wiphy_list, wiphy_match, L_UINT_TO_PTR(id));
-	if (!wiphy)
-		return;
+	/*
+	 * TODO: Support at least Link Measurement if TX_POWER_INSERTION is
+	 * available
+	 */
+}
+
+void wiphy_create_complete(struct wiphy *wiphy)
+{
+	if (util_mem_is_zero(wiphy->permanent_addr, 6)) {
+		int err = wiphy_get_permanent_addr_from_sysfs(wiphy);
+
+		if (err < 0)
+			l_error("Can't read sysfs maccaddr for %s: %s",
+					wiphy->name, strerror(-err));
+	}
+
+	wiphy_set_station_capability_bits(wiphy);
+	wiphy_setup_rm_enabled_capabilities(wiphy);
+
+	wiphy_print_basic_info(wiphy);
+}
+
+bool wiphy_destroy(struct wiphy *wiphy)
+{
+	l_debug("");
+
+	if (!l_queue_remove(wiphy_list, wiphy))
+		return false;
 
 	l_dbus_unregister_object(dbus_get_bus(), wiphy_get_path(wiphy));
 
 	wiphy_free(wiphy);
-}
-
-static void wiphy_config_notify(struct l_genl_msg *msg, void *user_data)
-{
-	uint8_t cmd;
-
-	cmd = l_genl_msg_get_command(msg);
-
-	l_debug("Notification of command %u", cmd);
-
-	switch (cmd) {
-	case NL80211_CMD_NEW_WIPHY:
-		wiphy_new_wiphy_event(msg);
-		break;
-	case NL80211_CMD_DEL_WIPHY:
-		wiphy_del_wiphy_event(msg);
-		break;
-	}
-}
-
-static void wiphy_regulatory_notify(struct l_genl_msg *msg, void *user_data)
-{
-	struct l_genl_attr attr;
-	uint16_t type, len;
-	const void *data;
-	uint8_t cmd;
-
-	cmd = l_genl_msg_get_command(msg);
-
-	l_debug("Regulatory notification %u", cmd);
-
-	if (!l_genl_attr_init(&attr, msg))
-		return;
-
-	while (l_genl_attr_next(&attr, &type, &len, &data)) {
-	}
-}
-
-static void regulatory_info_callback(struct l_genl_msg *msg, void *user_data)
-{
-	struct l_genl_attr attr;
-	uint16_t type, len;
-	const void *data;
-
-	if (!l_genl_attr_init(&attr, msg))
-		return;
-
-	while (l_genl_attr_next(&attr, &type, &len, &data)) {
-		switch (type) {
-		case NL80211_ATTR_REG_ALPHA2:
-			if (len != 3) {
-				l_warn("Invalid regulatory alpha2 attribute");
-				return;
-			}
-
-			l_debug("Regulatory alpha2 is %s", (char *) data);
-			break;
-		}
-	}
-}
-
-static void protocol_features_callback(struct l_genl_msg *msg, void *user_data)
-{
-	struct l_genl_attr attr;
-	uint16_t type, len;
-	const void *data;
-	uint32_t features = 0;
-
-	if (!l_genl_attr_init(&attr, msg))
-		return;
-
-	while (l_genl_attr_next(&attr, &type, &len, &data)) {
-		switch (type) {
-		case NL80211_ATTR_PROTOCOL_FEATURES:
-			if (len != sizeof(uint32_t)) {
-				l_warn("Invalid protocol features attribute");
-				return;
-			}
-
-			features = *((uint32_t *) data);
-			break;
-		}
-	}
-
-	if (features & NL80211_PROTOCOL_FEATURE_SPLIT_WIPHY_DUMP)
-		l_debug("Found split wiphy dump support");
+	return true;
 }
 
 static void wiphy_rfkill_cb(unsigned int wiphy_id, bool soft, bool hard,
@@ -792,6 +1003,7 @@ static void wiphy_rfkill_cb(unsigned int wiphy_id, bool soft, bool hard,
 	struct wiphy *wiphy = wiphy_find(wiphy_id);
 	struct l_dbus *dbus = dbus_get_bus();
 	bool old_powered, new_powered;
+	enum wiphy_state_watch_event event;
 
 	if (!wiphy)
 		return;
@@ -803,8 +1015,15 @@ static void wiphy_rfkill_cb(unsigned int wiphy_id, bool soft, bool hard,
 
 	new_powered = !wiphy->soft_rfkill && !wiphy->hard_rfkill;
 
-	if (old_powered != new_powered)
-		l_dbus_property_changed(dbus, wiphy_get_path(wiphy),
+	if (old_powered == new_powered)
+		return;
+
+	event = new_powered ? WIPHY_STATE_WATCH_EVENT_POWERED :
+				WIPHY_STATE_WATCH_EVENT_RFKILLED;
+	WATCHLIST_NOTIFY(&wiphy->state_watches, wiphy_state_watch_func_t,
+				wiphy, event);
+
+	l_dbus_property_changed(dbus, wiphy_get_path(wiphy),
 					IWD_WIPHY_INTERFACE, "Powered");
 }
 
@@ -905,6 +1124,11 @@ static bool wiphy_property_get_name(struct l_dbus *dbus,
 	return true;
 }
 
+#define WIPHY_MODE_MASK	( \
+	(1 << (NL80211_IFTYPE_STATION - 1)) | \
+	(1 << (NL80211_IFTYPE_AP - 1)) | \
+	(1 << (NL80211_IFTYPE_ADHOC - 1)))
+
 static bool wiphy_property_get_supported_modes(struct l_dbus *dbus,
 					struct l_dbus_message *message,
 					struct l_dbus_message_builder *builder,
@@ -912,7 +1136,7 @@ static bool wiphy_property_get_supported_modes(struct l_dbus *dbus,
 {
 	struct wiphy *wiphy = user_data;
 	unsigned int j = 0;
-	char **iftypes = wiphy_get_supported_iftypes(wiphy);
+	char **iftypes = wiphy_get_supported_iftypes(wiphy, WIPHY_MODE_MASK);
 
 	l_dbus_message_builder_enter_array(builder, "s");
 
@@ -941,10 +1165,15 @@ static void setup_wiphy_interface(struct l_dbus_interface *interface)
 					NULL);
 }
 
-bool wiphy_init(struct l_genl_family *in, const char *whitelist,
-							const char *blacklist)
+static int wiphy_init(void)
 {
-	struct l_genl_msg *msg;
+	struct l_genl *genl = iwd_get_genl();
+	const struct l_settings *config = iwd_get_config();
+	const char *whitelist = iwd_get_phy_whitelist();
+	const char *blacklist = iwd_get_phy_blacklist();
+	const char *s;
+
+	nl80211 = l_genl_family_new(genl, NL80211_GENL_NAME);
 
 	/*
 	 * This is an extra sanity check so that no memory is leaked
@@ -955,33 +1184,8 @@ bool wiphy_init(struct l_genl_family *in, const char *whitelist,
 		l_queue_destroy(wiphy_list, NULL);
 	}
 
-	nl80211 = in;
-
-	if (!l_genl_family_register(nl80211, "config", wiphy_config_notify,
-								NULL, NULL))
-		l_error("Registering for config notification failed");
-
-	if (!l_genl_family_register(nl80211, "regulatory",
-					wiphy_regulatory_notify, NULL, NULL))
-		l_error("Registering for regulatory notification failed");
-
 	wiphy_list = l_queue_new();
 
-	msg = l_genl_msg_new(NL80211_CMD_GET_PROTOCOL_FEATURES);
-	if (!l_genl_family_send(nl80211, msg, protocol_features_callback,
-								NULL, NULL))
-		l_error("Getting protocol features failed");
-
-	msg = l_genl_msg_new(NL80211_CMD_GET_REG);
-	if (!l_genl_family_send(nl80211, msg, regulatory_info_callback,
-								NULL, NULL))
-		l_error("Getting regulatory info failed");
-
-	msg = l_genl_msg_new(NL80211_CMD_GET_WIPHY);
-	if (!l_genl_family_dump(nl80211, msg, wiphy_dump_callback,
-						NULL, wiphy_dump_done))
-		l_error("Getting all wiphy devices failed");
-
 	rfkill_watch_add(wiphy_rfkill_cb, NULL);
 
 	if (!l_dbus_register_interface(dbus_get_bus(),
@@ -999,10 +1203,22 @@ bool wiphy_init(struct l_genl_family *in, const char *whitelist,
 	if (blacklist)
 		blacklist_filter = l_strsplit(blacklist, ',');
 
-	return true;
+	s = l_settings_get_value(config, "General",
+						"AddressRandomizationRange");
+	if (s) {
+		if (!strcmp(s, "nic"))
+			mac_randomize_bytes = 3;
+		else if (!strcmp(s, "full"))
+			mac_randomize_bytes = 6;
+		else
+			l_warn("Invalid [General].AddressRandomizationRange"
+				" value: %s", s);
+	}
+
+	return 0;
 }
 
-bool wiphy_exit(void)
+static void wiphy_exit(void)
 {
 	l_strfreev(whitelist_filter);
 	l_strfreev(blacklist_filter);
@@ -1010,11 +1226,14 @@ bool wiphy_exit(void)
 	l_queue_destroy(wiphy_list, wiphy_free);
 	wiphy_list = NULL;
 
+	l_genl_family_free(nl80211);
 	nl80211 = NULL;
+	mac_randomize_bytes = 6;
 
 	l_dbus_unregister_interface(dbus_get_bus(), IWD_WIPHY_INTERFACE);
 
 	l_hwdb_unref(hwdb);
-
-	return true;
 }
+
+IWD_MODULE(wiphy, wiphy_init, wiphy_exit);
+IWD_MODULE_DEPENDS(wiphy, rfkill);
